@@ -1,5 +1,9 @@
 import os
 
+from dotenv import load_dotenv
+
+load_dotenv()
+
 VEMBDA_PUBLIC_KEY = """
 -----BEGIN PUBLIC KEY-----
 MIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEA3VrSo+PIYLsKaLwp6Kuw
@@ -17,6 +21,14 @@ czy4Vhr2WxlC6p9ekb3PY/8CAwEAAQ==
 -----END PUBLIC KEY-----
 """
 
+NAMESPACE = os.getenv("NAMESPACE")
+IS_VPC = os.getenv("IS_VPC", "").lower() == "true"
+MEMORY_LIMIT_MB = int(os.getenv("MEMORY_LIMIT_MB", "2048"))
+PORT = os.getenv("PORT", "8000")
+VELLUM_API_URL_HOST = os.getenv("VELLUM_API_URL_HOST", "localhost")
+VELLUM_API_URL_PORT = os.getenv("VELLUM_API_URL_PORT", 8000)
+CONCURRENCY = int(os.getenv("CONCURRENCY", "8"))
+
 
 def is_development() -> bool:
     return os.getenv("FLASK_ENV", "local") == "local"
