@@ -1,10 +1,9 @@
-from contextlib import redirect_stdout
 from datetime import datetime
 import importlib
 from io import StringIO
 import json
 import logging
-from multiprocessing import Manager, Process, Queue
+from multiprocessing import Process, Queue
 import os
 import random
 import string
@@ -12,8 +11,8 @@ import sys
 from threading import Event as ThreadingEvent
 import time
 from traceback import format_exc
-from uuid import uuid4
-from typing import TYPE_CHECKING, Any, Callable, Generator, Iterator, Optional, Tuple
+from uuid import UUID, uuid4
+from typing import Any, Callable, Generator, Iterator, Optional, Tuple, Type
 
 from pebble import concurrent
 from vellum_ee.workflows.display.workflows import BaseWorkflowDisplay
@@ -21,135 +20,36 @@ from vellum_ee.workflows.server.virtual_file_loader import VirtualFileFinder
 
 from vellum import Vellum, VellumEnvironment
 from vellum.workflows import BaseWorkflow
-from vellum.workflows.errors import WorkflowError, WorkflowErrorCode
 from vellum.workflows.events.types import BaseEvent
+from vellum.workflows.events.workflow import WorkflowEventDisplayContext
+from vellum.workflows.exceptions import WorkflowInitializationException
 from vellum.workflows.inputs import BaseInputs
 from vellum.workflows.nodes import BaseNode
+from vellum.workflows.nodes.mocks import MockNodeExecution
 from vellum.workflows.state.base import BaseState, StateMeta
 from vellum.workflows.state.context import WorkflowContext
+from vellum.workflows.state.store import EmptyStore
 from vellum.workflows.workflows.event_filters import all_workflow_event_filter
+from workflow_server.config import IS_VPC, VELLUM_API_URL_HOST, VELLUM_API_URL_PORT
 from workflow_server.core.cancel_workflow import CancelWorkflowWatcherThread
+from workflow_server.core.events import (
+    SPAN_ID_EVENT,
+    STREAM_FINISHED_EVENT,
+    VembdaExecutionFulfilledBody,
+    VembdaExecutionFulfilledEvent,
+)
+from workflow_server.core.utils import serialize_vembda_rejected_event
 from workflow_server.core.workflow_executor_context import (
     DEFAULT_TIMEOUT_SECONDS,
     BaseExecutorContext,
     NodeExecutorContext,
     WorkflowExecutorContext,
 )
-from workflow_server.utils.exceptions import WorkflowInitializationException
 from workflow_server.utils.log_proxy import redirect_log
-from workflow_server.utils.utils import (
-    VembdaExecutionFulfilledBody,
-    VembdaExecutionFulfilledEvent,
-    VembdaExecutionInitiatedBody,
-    VembdaExecutionInitiatedEvent,
-    get_version,
-)
-
-if TYPE_CHECKING:
-    from vellum_ee.workflows.display.types import WorkflowEventDisplayContext
 
 logger = logging.getLogger(__name__)
 
 
-@concurrent.process(timeout=DEFAULT_TIMEOUT_SECONDS)
-def execute_workflow_pebble_timeout(executor_context: WorkflowExecutorContext) -> dict:
-    return execute_workflow(executor_context, output={})
-
-
-def execute_workflow_process_timeout(executor_context: WorkflowExecutorContext) -> dict:
-    try:
-        with Manager() as manager:
-            output = manager.dict()
-
-            p = Process(
-                target=execute_workflow,
-                args=(
-                    executor_context,
-                    output,
-                ),
-            )
-            p.start()
-            p.join(executor_context.timeout)
-            if p.is_alive():
-                p.kill()
-
-            if output.get("exit_code") is None:
-                return {
-                    "log": "",
-                    "result": {},
-                    "stderr": "",
-                    "exit_code": -1,
-                    "timed_out": True,
-                }
-
-            return {
-                "log": output["log"],
-                "result": output["result"],
-                "stderr": output["stderr"],
-                "exit_code": output["exit_code"],
-                "timed_out": False,
-            }
-
-    except Exception as e:
-        logger.exception(e)
-
-        return {
-            "result": {},
-            "stderr": format_exc(),
-            "exit_code": -1,
-            "timed_out": False,
-        }
-
-
-def execute_workflow(executor_context: WorkflowExecutorContext, output: dict) -> dict:
-    log_redirect = StringIO()
-
-    try:
-        # We need to namespace to avoid any collisions with the pebble managed process pool that may or may not
-        # reuse processes and also this avoids collisions with local packages and also fixes the virtual loader
-        # not working with relative import paths. We also do this renaming on the fly in the virtual file finder
-        # to avoid copying the files dict and costing extra latency and ram.
-        namespace = _get_file_namespace(executor_context)
-        sys.meta_path.append(VirtualFileFinder(executor_context.files, namespace))
-        inputs = importlib.import_module(f"{namespace}.inputs")
-
-        with redirect_stdout(log_redirect):
-            result = _create_workflow(
-                executor_context=executor_context,
-                namespace=namespace,
-            ).run(
-                inputs=inputs.Inputs(**executor_context.inputs),
-            )
-
-        log = log_redirect.getvalue()
-
-        exec_output = result
-        output["result"] = exec_output.model_dump(mode="json")
-        output["log"] = log
-        output["stderr"] = ""
-        output["exit_code"] = 0
-    except Exception:
-        output["result"] = ""
-        output["exit_code"] = -1
-        output["log"] = log_redirect.getvalue()
-        output["stderr"] = format_exc()
-
-    return output
-
-
-@concurrent.process(timeout=DEFAULT_TIMEOUT_SECONDS)
-# type ignore since pebble annotation changes return type
-def stream_workflow_pebble_timeout(
-    executor_context: WorkflowExecutorContext,
-    queue: Queue,
-) -> None:
-    _stream_workflow_wrapper(
-        executor_context=executor_context,
-        output={},
-        queue=queue,
-    )
-
-
 @concurrent.process(timeout=DEFAULT_TIMEOUT_SECONDS)
 # type ignore since pebble annotation changes return type
 def stream_node_pebble_timeout(
@@ -158,14 +58,13 @@ def stream_node_pebble_timeout(
 ) -> None:
     _stream_node_wrapper(
         executor_context=executor_context,
-        output={},
         queue=queue,
     )
 
 
-def _stream_node_wrapper(executor_context: NodeExecutorContext, output: dict, queue: Queue) -> None:
+def _stream_node_wrapper(executor_context: NodeExecutorContext, queue: Queue) -> None:
     try:
-        for event in stream_node(executor_context=executor_context, output=output):
+        for event in stream_node(executor_context=executor_context):
             queue.put(event)
     except Exception as e:
         logger.exception(e)
@@ -185,99 +84,47 @@ def _stream_node_wrapper(executor_context: NodeExecutorContext, output: dict, qu
         )
 
 
-def _stream_workflow_wrapper(executor_context: WorkflowExecutorContext, output: dict, queue: Queue) -> None:
+def _stream_workflow_wrapper(executor_context: WorkflowExecutorContext, queue: Queue) -> None:
+    span_id_emitted = False
     try:
-        for event in stream_workflow(executor_context=executor_context, output=output):
-            queue.put(event)
+        stream_iterator, span_id = stream_workflow(executor_context=executor_context)
+
+        queue.put(f"{SPAN_ID_EVENT}:{span_id}")
+        span_id_emitted = True
+
+        for event in stream_iterator:
+            queue.put(json.dumps(event))
+
     except WorkflowInitializationException as e:
-        # If we fail at initialization the initiated event won't be sent so we need to send it here
-        # or django event handler gets confused.
-        queue.put(
-            _dump_event(
-                event=VembdaExecutionInitiatedEvent(
-                    id=uuid4(),
-                    timestamp=datetime.now(),
-                    trace_id=executor_context.trace_id,
-                    span_id=executor_context.execution_id,
-                    body=VembdaExecutionInitiatedBody(
-                        sdk_version=get_version().get("sdk_version"),
-                        server_version=get_version().get("server_version"),
-                        display_context=None,
-                    ),
-                    parent=None,
-                ),
-                executor_context=executor_context,
-            )
-        )
+        if not span_id_emitted:
+            queue.put(f"{SPAN_ID_EVENT}:{uuid4()}")
 
-        queue.put(
-            VembdaExecutionFulfilledEvent(
-                id=uuid4(),
-                timestamp=datetime.now(),
-                trace_id=executor_context.trace_id,
-                span_id=executor_context.execution_id,
-                body=VembdaExecutionFulfilledBody(
-                    exit_code=-1,
-                    stderr=str(e),
-                    container_overhead_latency=executor_context.container_overhead_latency,
-                ),
-                parent=None,
-            ).model_dump(mode="json")
-        )
+        queue.put(serialize_vembda_rejected_event(executor_context, str(e)))
     except Exception as e:
+        if not span_id_emitted:
+            queue.put(f"{SPAN_ID_EVENT}:{uuid4()}")
+
         logger.exception(e)
-        queue.put(
-            VembdaExecutionFulfilledEvent(
-                id=uuid4(),
-                timestamp=datetime.now(),
-                trace_id=executor_context.trace_id,
-                span_id=executor_context.execution_id,
-                body=VembdaExecutionFulfilledBody(
-                    exit_code=-1,
-                    stderr="Internal Server Error",
-                    container_overhead_latency=executor_context.container_overhead_latency,
-                ),
-                parent=None,
-            ).model_dump(mode="json")
-        )
+        queue.put(serialize_vembda_rejected_event(executor_context, "Internal Server Error"))
+    queue.put(STREAM_FINISHED_EVENT)
+
+    exit(0)
 
 
-def stream_workflow_process_timeout(executor_context: WorkflowExecutorContext, queue: Queue) -> None:
-    try:
-        with Manager() as manager:
-            output = manager.dict()
-
-            p = Process(
-                target=_stream_workflow_wrapper,
-                args=(
-                    executor_context,
-                    output,
-                    queue,
-                ),
-            )
-            p.start()
-            p.join(executor_context.timeout)
-            if p.is_alive():
-                p.kill()
-
-            if output.get("exit_code") is None:
-                vembda_fulfilled_event = VembdaExecutionFulfilledEvent(
-                    id=uuid4(),
-                    timestamp=datetime.now(),
-                    trace_id=executor_context.trace_id,
-                    span_id=executor_context.execution_id,
-                    body=VembdaExecutionFulfilledBody(
-                        exit_code=-1,
-                        timed_out=True,
-                        container_overhead_latency=executor_context.container_overhead_latency,
-                    ),
-                    parent=None,
-                )
-                queue.put(vembda_fulfilled_event.model_dump(mode="json"))
-
-    except Exception as e:
-        logger.exception(e)
+def stream_workflow_process_timeout(
+    executor_context: WorkflowExecutorContext,
+    queue: Queue,
+) -> Process:
+    workflow_process = Process(
+        target=_stream_workflow_wrapper,
+        args=(
+            executor_context,
+            queue,
+        ),
+    )
+    workflow_process.start()
 
+    if workflow_process.exitcode is not None:
         vembda_fulfilled_event = VembdaExecutionFulfilledEvent(
             id=uuid4(),
             timestamp=datetime.now(),
@@ -285,98 +132,100 @@ def stream_workflow_process_timeout(executor_context: WorkflowExecutorContext, q
             span_id=executor_context.execution_id,
             body=VembdaExecutionFulfilledBody(
                 exit_code=-1,
-                stderr=format_exc(),
+                timed_out=True,
                 container_overhead_latency=executor_context.container_overhead_latency,
             ),
             parent=None,
         )
         queue.put(vembda_fulfilled_event.model_dump(mode="json"))
 
+    return workflow_process
+
 
 def stream_workflow(
     executor_context: WorkflowExecutorContext,
-    output: dict,
-    disable_redirect: bool = False,
-) -> Iterator[dict]:
+    disable_redirect: bool = True,
+) -> tuple[Iterator[dict], UUID]:
     workflow, namespace = _gather_workflow(executor_context)
     workflow_inputs = _get_workflow_inputs(executor_context)
+    display_context = _gather_display_context(workflow, namespace)
+    workflow_state = (
+        workflow.deserialize_state(
+            executor_context.state,
+            workflow_inputs=workflow_inputs or BaseInputs(),
+        )
+        if executor_context.state
+        else None
+    )
+    run_from_node = _get_run_from_node(executor_context, workflow)
+    node_output_mocks = MockNodeExecution.validate_all(
+        executor_context.node_output_mocks,
+        workflow.__class__,
+    )
+
+    cancel_watcher_kill_switch = ThreadingEvent()
+    cancel_signal = ThreadingEvent()
+    cancel_watcher = CancelWorkflowWatcherThread(
+        kill_switch=cancel_watcher_kill_switch,
+        execution_id=executor_context.execution_id,
+        timeout_seconds=executor_context.timeout,
+        vembda_public_url=executor_context.vembda_public_url,
+        cancel_signal=cancel_signal,
+    )
+
+    try:
+        if executor_context.vembda_public_url:
+            cancel_watcher.start()
+    except Exception:
+        logger.exception("Failed to start cancel watcher")
+
+    try:
+        stream = workflow.stream(
+            inputs=workflow_inputs,
+            state=workflow_state,
+            node_output_mocks=node_output_mocks,
+            event_filter=all_workflow_event_filter,
+            cancel_signal=cancel_signal,
+            entrypoint_nodes=[run_from_node] if run_from_node else None,
+        )
+    except Exception:
+        cancel_watcher_kill_switch.set()
+        logger.exception("Failed to generate Workflow Stream")
+        raise
 
     def call_workflow() -> Generator[dict[str, Any], Any, None]:
-        cancel_watcher_kill_switch = ThreadingEvent()
-        cancel_signal = ThreadingEvent()
-        cancel_watcher = CancelWorkflowWatcherThread(
-            kill_switch=cancel_watcher_kill_switch,
-            execution_id=executor_context.execution_id,
-            timeout_seconds=executor_context.timeout,
-            vembda_public_url=executor_context.vembda_public_url,
-            cancel_signal=cancel_signal,
-        )
-
         try:
-            if executor_context.vembda_public_url:
-                cancel_watcher.start()
-
-            try:
-                display_context = _gather_display_context(workflow, namespace)
-            except Exception:
-                logger.exception("Unable to Parse Workflow Display Context")
-                display_context = None
-
-            # This would be much more ideal if display_context lived in the workflow initiated event so that
-            # the vembda initiated could be fired off as soon as we receive the request.
-            yield _dump_event(
-                event=VembdaExecutionInitiatedEvent(
-                    id=uuid4(),
-                    timestamp=datetime.now(),
-                    trace_id=executor_context.trace_id,
-                    span_id=executor_context.execution_id,
-                    body=VembdaExecutionInitiatedBody(
-                        sdk_version=get_version().get("sdk_version"),
-                        server_version=get_version().get("server_version"),
-                        display_context=display_context,
-                    ),
-                    parent=None,
-                ),
-                executor_context=executor_context,
-            )
-
-            stream = workflow.stream(
-                inputs=workflow_inputs,
-                event_filter=all_workflow_event_filter,
-                cancel_signal=cancel_signal,
-            )
-
             first = True
             for event in stream:
                 if first:
                     executor_context.stream_start_time = time.time_ns()
                     first = False
-
-                if event.name.endswith("rejected") and event.body.error.code.name == "INTERNAL_ERROR":  # type: ignore
-                    logger.exception(event.body.error.message)  # type: ignore
-                    event.body.error = WorkflowError(  # type: ignore
-                        code=WorkflowErrorCode.INTERNAL_ERROR.value, message="Internal Error"  # type: ignore
-                    )
+                    if event.name == "workflow.execution.initiated":
+                        event.body.display_context = display_context
 
                 yield _dump_event(
                     event=event,
                     executor_context=executor_context,
                 )
+        except Exception as e:
+            logger.exception("Failed to generate event from Workflow Stream")
+            raise e
         finally:
             cancel_watcher_kill_switch.set()
 
-    return _call_stream(
-        executor_context=executor_context,
-        output=output,
-        stream_generator=call_workflow,
-        disable_redirect=disable_redirect,
+    return (
+        _call_stream(
+            executor_context=executor_context,
+            stream_generator=call_workflow,
+            disable_redirect=disable_redirect,
+        ),
+        stream.span_id,
     )
 
 
 def stream_node(
     executor_context: NodeExecutorContext,
-    output: dict,
-    disable_redirect: bool = False,
+    disable_redirect: bool = True,
 ) -> Iterator[dict]:
     namespace = _get_file_namespace(executor_context)
 
@@ -388,20 +237,10 @@ def stream_node(
         Node = getattr(node_module, executor_context.node_name)
 
         workflow_inputs = _get_workflow_inputs(executor_context)
-
-        State = (
-            importlib.import_module(f"{namespace}.state").State if executor_context.files.get("state.py") else BaseState
-        )
-
-        if executor_context.state:
-            state = State.parse_raw(executor_context.state)
-        else:
-            state = State(
-                meta=StateMeta(workflow_inputs=workflow_inputs),
-            )
+        workflow_state = _get_workflow_state(executor_context, workflow_inputs=workflow_inputs)
 
         node = Node(
-            state=state,
+            state=workflow_state,
             context=workflow_context,
         )
 
@@ -416,7 +255,6 @@ def stream_node(
 
     return _call_stream(
         executor_context=executor_context,
-        output=output,
         stream_generator=call_node,
         disable_redirect=disable_redirect,
     )
@@ -424,9 +262,8 @@ def stream_node(
 
 def _call_stream(
     executor_context: BaseExecutorContext,
-    output: dict,
     stream_generator: Callable[[], Generator[dict[str, Any], Any, None]],
-    disable_redirect: bool = False,
+    disable_redirect: bool = True,
 ) -> Iterator[dict]:
     log_redirect: Optional[StringIO] = None
 
@@ -436,7 +273,6 @@ def _call_stream(
     try:
         yield from stream_generator()
 
-        output["exit_code"] = 0
         vembda_fulfilled_event = VembdaExecutionFulfilledEvent(
             id=uuid4(),
             timestamp=datetime.now(),
@@ -453,7 +289,6 @@ def _call_stream(
         yield vembda_fulfilled_event.model_dump(mode="json")
 
     except Exception:
-        output["exit_code"] = -1
         vembda_fulfilled_event = VembdaExecutionFulfilledEvent(
             id=uuid4(),
             timestamp=datetime.now(),
@@ -482,12 +317,18 @@ def _create_workflow(executor_context: WorkflowExecutorContext, namespace: str)
         },
     )
 
-    return Workflow(context=workflow_context)
+    return Workflow(context=workflow_context, store=EmptyStore())
 
 
 def _create_workflow_context(executor_context: BaseExecutorContext) -> WorkflowContext:
-    if os.getenv("USE_LOCAL_VELLUM_API") == "true":
-        VELLUM_API_URL = "http://localhost:8000"
+    if IS_VPC:
+        environment = VellumEnvironment(
+            default=os.getenv("VELLUM_DEFAULT_API_URL", VellumEnvironment.PRODUCTION.default),
+            documents=os.getenv("VELLUM_DOCUMENTS_API_URL", VellumEnvironment.PRODUCTION.documents),
+            predict=os.getenv("VELLUM_PREDICT_API_URL", VellumEnvironment.PRODUCTION.predict),
+        )
+    elif os.getenv("USE_LOCAL_VELLUM_API") == "true":
+        VELLUM_API_URL = f"http://{VELLUM_API_URL_HOST}:{VELLUM_API_URL_PORT}"
         environment = VellumEnvironment(
             default=VELLUM_API_URL,
             documents=VELLUM_API_URL,
@@ -496,12 +337,16 @@ def _create_workflow_context(executor_context: BaseExecutorContext) -> WorkflowC
     else:
         environment = VellumEnvironment.PRODUCTION
 
+    if executor_context.environment_variables:
+        os.environ.update(executor_context.environment_variables)
+
     return WorkflowContext(
         vellum_client=Vellum(
-            api_key=executor_context.workspace_api_key,
+            api_key=executor_context.environment_api_key,
             environment=environment,
         ),
-        parent_context=executor_context.parent_context,
+        execution_context=executor_context.execution_context,
+        generated_files=executor_context.files,
     )
 
 
@@ -564,6 +409,40 @@ but found {inputs_module.Inputs.__class__.__name__}"""
     return inputs_module.Inputs(**executor_context.inputs)
 
 
+def _get_workflow_state(
+    executor_context: BaseExecutorContext, workflow_inputs: Optional[BaseInputs]
+) -> Optional[BaseState]:
+    namespace = _get_file_namespace(executor_context)
+    State = importlib.import_module(f"{namespace}.state").State if executor_context.files.get("state.py") else None
+
+    if not State:
+        return None
+
+    if not issubclass(State, BaseState):
+        return None
+
+    if executor_context.state:
+        return State(**executor_context.state)
+
+    if workflow_inputs:
+        return State(
+            meta=StateMeta(workflow_inputs=workflow_inputs),
+        )
+
+    return State()
+
+
+def _get_run_from_node(executor_context: WorkflowExecutorContext, workflow: BaseWorkflow) -> Optional[Type[BaseNode]]:
+    if not executor_context.node_id:
+        return None
+
+    for node in workflow.get_nodes():
+        if node.__id__ == executor_context.node_id:
+            return node
+
+    return None
+
+
 def _gather_workflow(context: WorkflowExecutorContext) -> Tuple[BaseWorkflow, str]:
     try:
         namespace = _get_file_namespace(context)
@@ -574,9 +453,13 @@ def _gather_workflow(context: WorkflowExecutorContext) -> Tuple[BaseWorkflow, st
         )
         return workflow, namespace
     except Exception as e:
+        logger.exception("Failed to initialize Workflow")
         raise WorkflowInitializationException(f"Failed to initialize workflow: {e}") from e
 
 
 def _gather_display_context(workflow: BaseWorkflow, namespace: str) -> Optional["WorkflowEventDisplayContext"]:
-    display_context = BaseWorkflowDisplay.gather_event_display_context(namespace, workflow.__class__)
-    return display_context
+    try:
+        return BaseWorkflowDisplay.gather_event_display_context(namespace, workflow.__class__)
+    except Exception:
+        logger.exception("Unable to Parse Workflow Display Context")
+        return None
